#include "stdinc.h"
#include "server.h"

#define NAME	"m_watch"
#define AUTHOR	"Jer"
#define VERSION "$Id: m_watch.c 1246 2008-08-17 05:11:02Z twitch $"
#define BUFSIZE	512

#define ACTION_NONE		0
#define ACTION_KICK		1
#define ACTION_KILL		2	
#define ACTION_GLINE	3

dlink_list	watchlist;
dlink_list	watchqueue;

struct watch_
{
	Channel		*channel;
	int		action;
	char		message[MAXLEN + 1];
	char		nickpattern[512];
};

struct watchqueue_
{
	User		*user;
	struct watch_	*watch;
};

struct watchaction_
{
	int	action;
	char	*name;
};

struct watchaction_ actions[15] = {
	{ ACTION_NONE,	"none"		},
	{ ACTION_KICK,	"kick"		},
	{ ACTION_KICK,	"kickban"	},
	{ ACTION_KILL,	"kill"		},
	{ ACTION_GLINE,	"gline"		},
	{ ACTION_GLINE,	"zline" 	},
	{ ACTION_GLINE, "gzline"	},
	{ ACTION_GLINE,	"kline"		},
	{ 0, NULL }
};

struct watch_	*is_watched (char *);

/* Event Handler for channel joins */
static void		evh_chanjoin    (int, evh_channel *);
static void		evh_user	(int, User *);
static void		evh_watchqueue	();

/* Our local routines */
static void		load_watchdb		(void);
static int		save_watchdb		(void);
static int		find_action		(char *);
static int		take_action		(User *, struct watch_ *);
static char		*action_num_to_name	(int);

/* Command handler */
static void		cmd_watch	(User *, char *);

/* Help Handlers */
static void		help_watch	(User *);
static void		help_watch_ext	(User *);

#ifdef HAVE_SQLITE
struct sqlite3	*watchdb;
static char		watchdb_file[MAXPATH];
#endif

int Module_Init()
{
	int	ch;

	RegisterModule (NAME,AUTHOR,VERSION,0,4,0);

#ifdef HAVE_SQLITE
	snprintf (watchdb_file, sizeof(watchdb_file), "%s/watchdb", datadir);

	if (s_Guardian)
	{
		ch = AddCmd (s_Guardian, "WATCH", ACC_OPER, cmd_watch);

		AddHelp    (s_Guardian, help_watch);
		AddHelpExt (ch, s_Guardian, help_watch_ext);
	}

	if (sqlite3_open (watchdb_file, &watchdb))
	{
		alog (3, "Unable to open watchdb");
		return MOD_CONT;
	}

	load_watchdb ();

	AddTimedEvent ("Watch DB Save", save_watchdb, 900);
	AddTimedEvent ("WatchQueue", evh_watchqueue, 30);
	AddEvent ("CHANJOIN", evh_chanjoin);
	AddEvent ("USERCONN", evh_user);
	return MOD_CONT;
#endif
	return MOD_ERR_SQL;
}

void Module_Close()
{

#ifdef HAVE_SQLITE
	DelTimedEvent ("Watch DB Save");
	DelTimedEvent ("WatchQueue");
	DelEvent ("CHANJOIN", evh_chanjoin);
	DelEvent ("USERCONN", evh_user);

	save_watchdb ();

	if (watchdb)
		sqlite3_close (watchdb);
	if (s_Guardian)
		DelCmd (s_Guardian, "WATCH", cmd_watch);
#endif

	return;
}

static void load_watchdb (void)
{
	Channel			*ch;
	dlink_node		*dl;
	struct watch_	*watch;
	char 			**data, *errmsg, *chname;
	int				i, j, row_count, col_count, pos;

	watchlist.head	= watchlist.tail	= NULL;

#ifdef HAVE_SQLITE

	if ((i = sqlite3_get_table (watchdb, "SELECT * FROM channels", &data, &row_count, &col_count, &errmsg)))
	{
		switch (i)
		{
			case SQLITE_BUSY:
				alog (3, "Unable to load channels from watch database, SQLite Busy");
				break;

			default:
				alog (3, "Problem with watchdb file: %s", errmsg);
				break;
		}

        ircd_wallop (s_Guardian->nick, "Error during watch database load");
        return;

	}

	if (row_count == 0)
	{
		return;
	}

	pos	= col_count;

	for (i = 0; i < row_count; i++)
	{
		watch	= (struct watch_ *) malloc (sizeof(struct watch_));
		dl		= dlink_create ();

		for (j = 0; j < col_count; j++)
		{
			switch (j)
			{
				case 0:
					chname	= data[pos++];

					if (!(ch = find_channel (chname)))
						ch	= new_chan (chname);

					watch->channel	= ch;

					if (!in_channel (s_Guardian, ch))
					{
						AddToChannelU (s_Guardian, ch);
						ircd_join (s_Guardian, ch);
					}

					break;
				case 1:
					watch->action	= atoi(data[pos++]);
					break;
				case 2:
					strlcpy (watch->message, data[pos++], sizeof(watch->message));
					break;
			}
		}

		dlink_add_tail (watch, dl, &watchlist);
	}

	if (data)
		free (data);

	row_count = 0;
	col_count = 0;

	if ((i = sqlite3_get_table (watchdb, "SELECT * FROM nicknames", &data, &row_count, &col_count, &errmsg)))
	{
		switch (i)
		{
			case SQLITE_BUSY:
				alog (3, "Unable to load nicknames from watch database, SQLite Busy");
				break;

			default:
				alog (3, "Problem with watchdb file: %s", errmsg);
				break;

		}

		ircd_wallop (s_Guardian->nick, "Error during watch database load");
		return;

	}

	if (row_count == 0)
		return;

	pos = col_count;

	for (i = 0; i < row_count; i++)
	{
		watch	= (struct watch_ *) malloc (sizeof(struct watch_));
		dl		= dlink_create ();

		for (j = 0; j < col_count; j++)
		{
			switch (j)
			{
				case 0:
					chname	= data[pos++];

					strlcpy (watch->nickpattern, chname, sizeof(watch->nickpattern));
					break;
				case 1:
					watch->action	= atoi(data[pos++]);
					break;
				case 2:
					strlcpy (watch->message, data[pos++], sizeof(watch->message));
					break;
			}
		}

		dlink_add_tail (watch, dl, &watchlist);
	}

	if (data)
		free (data);

#endif

	return;
}

static int save_watchdb (void)
{
	dlink_node		*dl;
	struct watch_	*watch;
	static char		buf[BUFSIZE * 4];
	char			*errmsg;
	int				i;

#ifdef HAVE_SQLITE

	alog (0, "Saving watchdb to %s", watchdb_file);

	if ((i = sqlite3_exec (watchdb, "DELETE FROM channels", NULL, NULL, &errmsg)))
	{

		switch (i)
		{
			case SQLITE_BUSY:
				alog (3, "Unable to clear channels table, SQLite Busy");
				break;

			default:
				alog (3, "Problem with watchdb file: %s", errmsg);
				break;
		}

		ircd_wallop (s_Guardian->nick, "Error during watch database save");
		return 0;
	}
	if ((i = sqlite3_exec (watchdb, "DELETE FROM nicknames", NULL, NULL, &errmsg)))
	{
		switch (i)
		{
			case SQLITE_BUSY:
				alog (3, "Unable to clear nicknames table, SQLite Busy");
				break;

			default:
				alog (3, "Problem with watchdb file: %s", errmsg);
				break;
		}

		ircd_wallop (s_Guardian->nick, "Error during watch database save");
		return 0;
	}

	DLINK_FOREACH (dl, watchlist.head)
	{
		watch	= (struct watch_ *) dl->data;

		if ((!watch) || (!watch->message) || (!watch->action))
			continue;

		if (watch->nickpattern[0] != '\0')
		{
			snprintf (buf, sizeof(buf), "INSERT INTO nicknames VALUES ('%s', %d, '%s')", watch->nickpattern, watch->action, watch->message);

			if ((i = sqlite3_exec (watchdb, buf, NULL, NULL, &errmsg)))
			{
				switch (i)
				{
					case SQLITE_BUSY:
						alog (3, "Unable to insert entry into nicknames table, SQLite Busy");
						break;

					default:
						alog (3, "Problem with watchdb file: %s", errmsg);
						break;
				}

				ircd_wallop (s_Guardian->nick, "Error during watch database save");
				return 0;
			}
		}
		else
		{
			snprintf (buf, sizeof(buf), "INSERT INTO channels VALUES ('%s', %d, '%s')", watch->channel->name, watch->action, watch->message);

			if ((i = sqlite3_exec (watchdb, buf, NULL, NULL, &errmsg)))
			{
				switch (i)
				{
					case SQLITE_BUSY:
						alog (3, "Unable to insert entry into channels table, SQLite Busy");
						break;

					default:
						alog (3, "Problem with watchdb file: %s", errmsg);
						break;
				}

				ircd_wallop (s_Guardian->nick, "Error during watch database save");
				return 0;
			}
		}
	}
#endif

	return 1;
}

static void cmd_watch (User *source, char *argv)
{
	int 	i, n;
	char	*params[512], message[1024];
	struct watch_	*watch;
	dlink_node		*dl, *tdl;
	Channel			*ch;

	n = generictoken (' ', sizeof(params), argv, params);

#ifdef HAVE_SQLITE

	if ((strcasecmp (params[0], "LIST")) == 0)
	{
		sendto_one (s_Guardian, source, "Displaying channels/nickname patterns we currently are watching");

		DLINK_FOREACH (dl, watchlist.head)
		{
			watch	= (struct watch_ *) dl->data;

			if (watch->nickpattern[0] != '\0')
				sendto_one (s_Guardian, source, "%s [Action: %s] [Message: %s]", watch->nickpattern, action_num_to_name(watch->action), watch->message);
			else
				sendto_one (s_Guardian, source, "%s [Action: %s] [Message: %s]", (watch->channel)? watch->channel->name : "No channel name", action_num_to_name(watch->action), watch->message);
		}

		return;
	}
	else if ((strcasecmp (params[0], "ADD")) == 0)
	{
		if (n < 4)
		{
			sendto_one (s_Guardian, source, "Syntax: ADD <channel|nickname> <action> <message>");
			return;
		}

		if (is_watched (params[1]))
		{
			sendto_one (s_Guardian, source, "%s is already being watched", params[1]);
			return;
		}

		watch	= (struct watch_ *) malloc (sizeof(struct watch_));
		dl		= dlink_create ();

        if (!isalpha(*params[2]))
        	watch->action   = atoi(params[2]);
        else
        	watch->action   = find_action (params[2]);

		memset (watch->nickpattern, '\0', sizeof(watch->nickpattern));

		if (*params[1] == '#')
		{

			if (!(ch = find_channel (params[1])))
				ch	= new_chan (params[1]);

			if (!ch)
			{
				free (watch);
				dlink_free (dl);
				return;
			}

			watch->channel	= ch;

                	if (!in_channel (s_Guardian, ch))
                	{
                        	AddToChannelU (s_Guardian, ch);

                        	if (sync_state != BURSTING)
                                	ircd_join (s_Guardian, ch);
                	}
		}
		else
		{
			strlcpy (watch->nickpattern, params[1], sizeof(watch->nickpattern));
			watch->channel	= NULL;

			if (watch->action == ACTION_KICK)
			{
				sendto_one (s_Guardian, source, "Nickname patterns cannot use actions kick/kickban");
				free (watch);
				dlink_free (dl);
				return;
			}

		}

		ExpandParv (message, sizeof(message), 3, n, params);

		strlcpy (watch->message, message, sizeof(watch->message));

		dlink_add_tail (watch, dl, &watchlist);

		sendto_one (s_Guardian, source, "Added watch for %s (Action: %s Message: %s)", (watch->nickpattern[0] == '\0')? ch->name : watch->nickpattern, action_num_to_name(watch->action), message);

		return;
	}
	else if ((strcasecmp (params[0], "REMOVE")) == 0)
	{
		if (n < 2)
			return;

		if (!is_watched (params[1]))
		{
			sendto_one (s_Guardian, source, "Channel %s is not being watched", params[1]);
			return;
		}

		DLINK_FOREACH_SAFE (dl, tdl, watchlist.head)
		{
			watch	= (struct watch_ *) dl->data;

			if (watch->channel)
			{

				if ((strcasecmp (watch->channel->name, params[1])) == 0)
				{
					if (in_channel (s_Guardian, watch->channel))
					{
						DelFromChannel (s_Guardian, watch->channel);

						send_line (":%s PART %s :Leaving", s_Guardian->nick, watch->channel->name);
					}

					dlink_delete (dl, &watchlist);
					dlink_free (dl);
					free (watch);

					sendto_one (s_Guardian, source, "No longer watching %s", params[1]);

					return;
				}
			}
			else if (watch->nickpattern[0] != '\0')
			{
				if ((strcasecmp (watch->nickpattern, params[1])) == 0)
				{
					dlink_delete (dl, &watchlist);
					dlink_free (dl);
					free (watch);

					sendto_one (s_Guardian, source, "No longer watching %s", params[1]);
					return;
				}
			}
		}
		return;
	}
	else if ((strcasecmp (params[0], "SAVE")) == 0)
	{
		save_watchdb ();

		sendto_one (s_Guardian, source, "Saved Watch Database");
		return;
	}
	else if ((strcasecmp (params[0], "LOAD")) == 0)
	{
		DLINK_FOREACH_SAFE (dl, tdl, watchlist.head)
		{
			watch	= (struct watch_ *) dl->data;

			dlink_delete (dl, &watchlist);
			dlink_free (dl);
			free (watch);
		}

		load_watchdb ();
		sendto_one (s_Guardian, source, "Loaded Watch Database");
		return;
	}
	else if ((strcasecmp (params[0], "CHANGE")) == 0)
	{
        if (n < 4)
        {
                sendto_one (s_Guardian, source, "Syntax: CHANGE <channel> <action|message> <value>");
                return;
        }

		if (!(watch = is_watched (params[1])))
		{
			sendto_one (s_Guardian, source, "%s is not currently being watched", params[1]);
			return;
		}

		if ((strcasecmp (params[2], "action")) == 0)
		{
			if (!isalpha(*params[3]))
				watch->action	= atol(params[3]);
			else
				watch->action	= find_action(params[3]);

			if ((watch->nickpattern[0] != '\0') && (watch->action == ACTION_KICK))
            {
            	sendto_one (s_Guardian, source, "Nickname patterns cannot use actions kick/kickban");
            	free (watch);
            	dlink_free (dl);
            	return;
            }

			sendto_one (s_Guardian, source, "Changed action to (%s) on watch %s", action_num_to_name(watch->action), (watch->channel)? watch->channel->name : watch->nickpattern);

			return;
		}
		else if ((strcasecmp (params[2], "message")) == 0)
		{
			ExpandParv (message, sizeof(message), 3, n, params);

			strlcpy (watch->message, message, sizeof(watch->message));

			sendto_one (s_Guardian, source, "Changed message to (%s) on watched channel %s", message, watch->channel->name);
			return;
		}
		else
		{
			sendto_one (s_Guardian, source, "Invalid change option: %s", params[2]);
			return;
		}

	}

#else

	sendto_one (s_Guardian, source, "SQLite support is not present, WATCH has been disabled.");

#endif
	
	return;
}

void evh_chanjoin (int argc, evh_channel *evhdata)
{
	struct watch_		*watch;
	struct watchqueue_	*watchque;
	dlink_node		*dl;
	char			mask[512];

	if ((MyConnect(evhdata->u)) || (evhdata->u->service))
		return;

	if ((watch = is_watched (evhdata->c->name)))
	{
		if (sync_state == BURSTING)
		{
			watchque	= (struct watchqueue_ *) malloc (sizeof(struct watchqueue_));
			dl		= dlink_create ();

			watchque->user	= evhdata->u;
			watchque->watch	= watch;

			dlink_add_tail (watchque, dl, &watchqueue);

			return;
		}

		sendto_logchan ("\002WATCH:\002 %s is joining a watched channel (%s)", evhdata->u->nick, evhdata->c->name);

		take_action (evhdata->u, watch);

		return;
	}
	return;
}

void evh_user (int argc, User *source)
{
	struct watch_		*watch;
	struct watchqueue_	*watchque;
	dlink_node		*dl, *ndl;
	int					status;

	if ((MyConnect(source)) || (source->service))
		return;

	DLINK_FOREACH (dl, watchlist.head)
	{
		watch	= (struct watch_ *) dl->data;

		if (watch->nickpattern[0] != '\0')
		{
			if ((status = match_regx (source->nick, watch->nickpattern)) == 0)
			{
				if (sync_state == BURSTING)
				{
					watchque	= (struct watchqueue_ *) malloc (sizeof(struct watchqueue_));
					ndl		= dlink_create ();

					watchque->user	= source;
					watchque->watch	= watch;

					dlink_add_tail (watchque, ndl, &watchqueue);
					return;
				}

				take_action (source, watch);
			}

		}
	}

	return;
}

void evh_watchqueue ()
{
	dlink_node		*dl, *tdl;
	struct watch_		*watch;
	struct watchqueue_	*watchque;
	User			*user;
	int			i;

	i = 0;

	if (sync_state == BURSTING)
		return;

	DLINK_FOREACH_SAFE (dl, tdl, watchqueue.head)
	{
		if (i > 25)
			return;

		i++;

		watchque	= (struct watchqueue_ *) dl->data;
		watch		= watchque->watch;
		user		= watchque->user;

		if (watch->nickpattern[0] != '\0')
			sendto_logchan ("\002WATCH:\002 %s joined the network during burst (%s)\n", user->nick, watch->nickpattern);
		else
			sendto_logchan ("\002WATCH:\002 %s joined a watched channel during burst (%s)\n", user->nick, watch->channel->name);

		take_action (user, watch);


		dlink_delete (dl, &watchqueue);
		dlink_free (dl);
		free (watchque);
	}

	DelTimedEvent ("WatchQueue");
	return;
}

struct watch_ *is_watched (char *ch)
{
	dlink_node		*dl;
	struct watch_	*watch;

	if (!ch)
		return 0;

	DLINK_FOREACH (dl, watchlist.head)
	{
		watch	= (struct watch_ *) dl->data;

		if (watch->nickpattern[0] != '\0')
		{
			if ((strcasecmp (watch->nickpattern, ch)) == 0)
				return watch;

			continue;
		}

		if (!watch->channel)
			continue;

		if ((strcasecmp (watch->channel->name, ch)) == 0)
			return watch;
	}

	return NULL;
}

static int take_action	(User *source, struct watch_ *watch)
{
	if (!source || !watch)
		return -1;

	switch (watch->action)
	{
		case ACTION_NONE:
			break;
		case ACTION_KICK:
			send_line (":%s MODE %s +b *!*@%s", (IRCd.ts6)? s_Guardian->uid : s_Guardian->nick, watch->channel->name, source->host);
			send_line (":%s KICK %s %s :%s", (IRCd.ts6)? s_Guardian->uid : s_Guardian->nick, watch->channel->name, source->nick, watch->message);

			DelFromChannel (source, watch->channel);
			break;
		case ACTION_KILL:
			ircd_kill (source, watch->message);
			break;
		case ACTION_GLINE:
			ircd_kill (source, watch->message);
			ircd_xline ("G", (IRCd.ts6)? s_Guardian->uid : s_Guardian->nick, source->host, watch->message);
			break;
	}

	if (watch->action == ACTION_NONE)
		sendto_logchan ("\002WATCH:\002 no action taken on %s", source->nick);
	else
		sendto_logchan ("\002WATCH:\002 action %s taken on %s, matched watch rule %s", action_num_to_name(watch->action), source->nick,
					(watch->channel)? watch->channel->name : watch->nickpattern);

	return 0;
}

static int find_action	(char *name)
{
	int	i;

	for (i = 0; actions[i].name != NULL; i++)
	{
		if ((strcasecmp (actions[i].name, name)) == 0)
			return actions[i].action;
	}

	return 0;
}

static char *action_num_to_name (int action)
{
	int i;

	for (i = 0; actions[i].name != NULL; i++)
	{
		if (actions[i].action == action)
			return actions[i].name;
	}

	return "Unknown";
}

static void help_watch	(User *source)
{
	sendto_one (s_Guardian, source, "%s %-20s", "WATCH", " - Manage the channel/nick watch list");
	return;
}

static void help_watch_ext	(User *source)
{
	sendto_one (s_Guardian, source, "Syntax \002WATCH [OPTION]\002");

	sendto_one (s_Guardian, source, " ");

	sendto_one (s_Guardian, source, "\002WATCH LIST\002");
	sendto_one (s_Guardian, source, " ");
	sendto_one (s_Guardian, source, "    Display channels/nicknames we are currently watching");
	sendto_one (s_Guardian, source, " ");

	sendto_one (s_Guardian, source, "\002WATCH ADD [channel|nickname] [action] [message]");
	sendto_one (s_Guardian, source, " ");
	sendto_one (s_Guardian, source, "    Add the given channel or nickname to our watch list");
	sendto_one (s_Guardian, source, " ");
	sendto_one (s_Guardian, source, "    Valid actions: none,kick,kickban,kill,gline,kline,gzline,zline");
	sendto_one (s_Guardian, source, " ");

	sendto_one (s_Guardian, source, "\002WATCH REMOVE [channel|nickname]\002");
	sendto_one (s_Guardian, source, " ");
	sendto_one (s_Guardian, source, "    Remove the given channel or nickname from the watch list");
	sendto_one (s_Guardian, source, " ");

	sendto_one (s_Guardian, source, "\002WATCH CHANGE [channel|nickname] [action|message] [value]\002");
	sendto_one (s_Guardian, source, " ");
	sendto_one (s_Guardian, source, "    Change a setting on a previously set watch");
	sendto_one (s_Guardian, source, " ");

	sendto_one (s_Guardian, source, "\002WATCH SAVE\002");
	sendto_one (s_Guardian, source, " ");
	sendto_one (s_Guardian, source, "    Save the watch list to the database");
	sendto_one (s_Guardian, source, " ");

	sendto_one (s_Guardian, source, "\002WATCH LOAD\002");
	sendto_one (s_Guardian, source, " ");
	sendto_one (s_Guardian, source, "    Load the watch list from the database");
	sendto_one (s_Guardian, source, " ");

	return;
}


