/*
 *          OMEGA SECURITY SERVICES
 *       (C) 2008 Omega Development Team
 *
 * Name:  access.c
 * Copyright:  2008 Omega Security Services
 * Author:  Omega Dev Team
 * Description: access management and routines
 *
 *          $Id: $
 */


#include "stdinc.h"
#include "server.h"



/*******************************************/
/**
 * find_access() - Find the FIRST matching access
 * access list entry for the given user and return
 * the access list entry.
 *
 * @param u - Pointer to the user struct
 *            that we are checking with.
 *
 * @return Access* - return a pointer to the
 *                   access entry list that the user
 *                   that the user belongs to, or
 *                   NULL if no match is found.
 *
 */


Access* find_access(User *u)
{
	dlink_node *dl,*hdl;
	char *host;
	Access *a;
	char pattern[150];

    sprintf(pattern,"%s@%s",u->user,u->host);

	DLINK_FOREACH(dl,accesslist.head)
	{
		a = dl->data;


		DLINK_FOREACH(hdl,a->hosts.head)
		{
		 	host = hdl->data;

			if (match_wild(host,pattern,0))
				return a;
		}
	}
	return NULL;

}


/*******************************************/

int check_access(User* u, uint32 level)
{

	if (!u)
		return 0;

	if (!level)
		return 1;

	//user has no access
	if (!u->access)
		return 0;

	if (u->access & level)
		return 1;

	return 0;


}

/*******************************************/
/**
 * setAccess() - finds then sets or removes
 * access for a user, this will set the first
 * matching access level on a user, as provided '
 * by find_access().
 *
 *  @param u - The user struct we are working with
 *
 *  @param op - wether or not we are adding/removing
 *              access from the user struct.
 *
 *  @return int - Returns true if access has been set
 *                or false if no record was found.
 */


int setAccess(User* u, int op)
{

	Access *a;

	if (!u)
		return 0;

	if (op)
	{
		if (!(a = find_access(u)))
		{
			u->access = ACC_OPER;
			return 0;
		}

		alog(DEBUG3,"Setting access for %s (Access: %lu)",u->nick,(unsigned long) a->flags);
        u->access = a->flags;
		return 1;
	}

	alog(DEBUG3,"Removing access for %s",u->nick);
	u->access &= ~u->access;
	return 1;
}

/*************************************************/
/**
 *  Destroy the access list
 *   This function is primarily used on rehash and shutdown
 *   when its nessicary to start fresh and recreate a new list.
 *
 */

int DestAccList()
{
	dlink_node *dl;
	dlink_node *tdl;

	dlink_node *hdl,*thdl; //host dlink list
	char *host;

	Access *a;

	DLINK_FOREACH_SAFE(dl,tdl,accesslist.head)
	{

		a = dl->data;

		DLINK_FOREACH_SAFE(hdl,thdl,a->hosts.head)
		{
			host = hdl->data;

			dlink_find_delete(host,&a->hosts);
			dlink_free(hdl);
			free(host);
		}


		dlink_find_delete(a,&accesslist);
		dlink_free(dl);
		free(a);

	}
	return 0;

}

/*************************************************/
/**
 *   refresh_access() - Crawls our access list
 *   finds an entry then craws the user list and
 *   reassigns users access - this helps us with
 *   stale entries.
 *
 *   NOTE: Since this is a slow method
 *   of doing this, recommend only doing this
 *   if our list was destroyed such as rehash
 */


void refresh_access(void)
{
    User* u = NULL;
    Access* a = NULL;

    dlink_node *udl = NULL;


     DLINK_FOREACH(udl,userlist.head)
     {
                u = udl->data;

		/**
		 * if there not an OPER or they are a SERVICE
         * they have no access. so skip that entry :)
         * this should give us back some speed and
         * proformance here. Esp on networks with more
         * Services and Users then IRCops (Which should
         * be all of them imo)
         * - Twitch
		 */

		if ((!u->oper) || (u->service))
			continue;

		setAccess(u,1);
     }

    return;
}

/* EOF */

