#include "stdinc.h"
#include "server.h"
#include "extern.h"

extern Events* event_list;
extern Module* currentModule;

dlink_list servcmds;
dlink_list usercmds;

static void SockEventNull(Socket*,int);
SockEvent	*FindSockEvent (char *, void (*)());

/*****************************************/

/**
 * add_sock_handlers - Adds a read/write handler to the socket
 *                     this is in hopes to get rid of HandleSockEvent
 *                     Which cuases alot of un-nessicary calls when threaded
 *                     realisticly we only want 1 handler per event, per socket.
 *
 * 	void	(*read_callback)  (struct socket_ *, int);
 *	void	(*write_callback) (void *);
 *
 */


void add_sock_handlers(Socket *s, void (*read)(Socket*, int),void (*write)(void*))
{
    if (!s || !read)
    {
        alog(2, "No read handler specified when adding sock handlers");
        return;
    }

    s->read_callback = read;

    if (write)
        s->write_callback = write;
}

/*****************************************/

void AddEvent(char *event, void* fnct)
{
    Events      *tmp;
    dlink_node  *dl;

    tmp = (Events*)malloc(sizeof(Events));

    if (!tmp)
    {
	alog(3,"Unable to add event %s malloc failed",event);
        return;
    }

    memcounts.eventssize    += sizeof(tmp);
    memcounts.eventscnt++;

    strlcpy (tmp->name, event, sizeof(tmp->name));

    tmp->handler = fnct;
    tmp->age = time(NULL);
    tmp->interval   = -1;

    dl  = dlink_create ();

    dlink_add_tail (tmp, dl, &events);

    return;
}

void DelEvent(char *event, void* fnct)
{

       Events      *tmp;
       dlink_node  *dl;


	DLINK_FOREACH(dl, events.head)
    	{
		tmp = dl->data;
		if (tmp->handler != fnct)
			continue;

	     dlink_delete (dl, &events);
            dlink_free (dl);
            free (tmp);
            return;

	}


	return;
}

void AddTimedEvent (char *event, void *fnct, int interval)
{
    Events      *tmp;
    dlink_node  *dl;

    tmp = (Events *) malloc (sizeof(Events));

    if (!tmp)
    {
        alog (ERROR, "%s: Unable to malloc()",__PRETTY_FUNCTION__);
        return;
    }

    memcounts.eventssize    += sizeof(tmp);
    memcounts.eventscnt++;

    strlcpy (tmp->name, event, sizeof(tmp->name));

    tmp->handler    = fnct;
    tmp->age        = time(NULL);
    tmp->interval   = interval;
    tmp->when       = time(NULL) + interval;

    dl  = dlink_create ();

    dlink_add_tail (tmp, dl, &events);
    return;
}

void RunTimedEvents ()
{
    dlink_node  *dl, *tdl;
    Events      *ev;

    DLINK_FOREACH(dl, events.head)
    {
        ev  = dl->data;

        if (ev->interval != -1)
        {
            if (ev->when <= ServerTime)
            {
                ev->when    = ServerTime + ev->interval;
                ev->handler (0, NULL); // Timed events do not need parameters
            }
        }
    }

    return;
}

void DelTimedEvent (char *event)
{
    dlink_node  *dl, *tdl;
    Events      *ev;

    DLINK_FOREACH(dl, events.head)
    {
        ev  = dl->data;

        if ((strcasecmp (ev->name, event)) == 0)
        {
            memcounts.eventssize    -= sizeof(ev);
            memcounts.eventscnt--;

            dlink_delete (dl, &events);
            dlink_free (dl);
            free (ev);
            return;
        }
    }
    return;
}



/*****************************************/

#if 0
void Event(char *event, int argc, char **argv)
#endif
void Event(char *event, int argc, void *argv)
{
    Events      *ev;
    dlink_node  *dl;

    if (debug)
	alog (5, "Emitting Event %s", event);

    DLINK_FOREACH (dl, events.head)
    {
        ev  = dl->data;

        if ((strcasecmp (ev->name, event)) == 0)
            ev->handler (argc, argv);
    }

	return;
}

/*****************************************/

void AddServCmd(char *cmd,void (*fnct)(Link*,int,char**))
{

    dlink_node *dl  = dlink_create ();
    ServCmds *c = (ServCmds*) malloc(sizeof(ServCmds));
    int index = HASH(cmd,1024);

    strcpy(c->name, cmd);
    c->fnct = fnct;

    dlink_add_tail(c, dl, &ServCmdTable[index]);

}

/*****************************************/

int HandleServCmd(char *command, Link* li, int arc, char **arv)
{
    ServCmds *c;
    dlink_node *dl;
    int found = 0;
    int index = HASH(command,1024);

    DLINK_FOREACH(dl,ServCmdTable[index].head)
    {
            c = dl->data;

            if (strcasecmp(c->name, command)==0)
            {
		    found = 1;
            	    c->fnct(li,arc,arv);
	     }
    }

    if (!found)
		alog(DEBUG3,"Handling unknown server command %s",command);


    return 0;

}

/**********************************************/

void DelServCmd(char *cmd,void (*fnct)(Link*,int,char**))
{
        ServCmds* c;
        dlink_node* dl;

        int index = HASH(cmd,1024);

        DLINK_FOREACH(dl,ServCmdTable[index].head)
        {
            c = dl->data;

            if (c->fnct == fnct)
            {

                    dl = dlink_find_delete(c,&UserCmdTable[index]);
                    dlink_free(dl);

                    free(c->module);
                    free(c);

                    return;

            }

        }

    return;
}



/**********************************************/
/**
 * Okay instead of using a fat dlinked list
 * index the commands by name into a hash
 * then retrieve assocated lists
 * this will save time in the future.
 */


void AddUserCmd(char *cmd,void (*fnct)(User*,int,char**))
{

    UserCmds* uc;
    dlink_node *dl;
    int index = HASH(cmd,1024);

    dl = dlink_create();

    if (!(uc = (UserCmds*) malloc(sizeof(UserCmds))))
        return;

    uc->routine = fnct;
    uc->active = 1;


    dlink_add_tail(uc,dl,&UserCmdTable[index]);

    return;

}

void DelUserCmd(char *cmd,void (*fnct)(User*,int,char**))
{
    UserCmds* uc;
    dlink_node* dl;

    int index = HASH(cmd,1024);

    DLINK_FOREACH(dl,UserCmdTable[index].head)
    {
            uc = dl->data;

            if (uc->routine == fnct)
            {
		    alog(DEBUG3,"Removing UserCmd associated with %s (Location: %p)",uc->module,uc->routine);
                    dl = dlink_find_delete(uc,&UserCmdTable[index]);
                    dlink_free(dl);

                    uc->active = 0;

                    free(uc->module);
                    free(uc);

                    return;

            }
    }


}



/*****************************************/

int HandleUserCmd(char *command, User *u, int arc, char **arv)
{

    int rc = 1;
	int index;

    UserCmds* uc;
    dlink_node* dl;

    if (!u || arc <= 0 || !arv)
        return 0;

    index = HASH(command,1024);

    DLINK_FOREACH(dl,UserCmdTable[index].head)
    {
            uc = dl->data;

	     assert(uc);

	     if (!uc)
		  continue;

	     if (!uc->routine)
			return 0;

	    alog(DEBUG3,"CMD: Calling USER Cmd: %p",uc->routine);

            uc->routine(u,arc,arv);
    }


	return 0;
}


/**********************************************/

void AddModuleMessage(char *msg, void (*fnct)  (char *src,char *dest,char *msg))
{
    ModuleMessage*  tmp;

    tmp = (ModuleMessage*)malloc(sizeof(ModuleMessage));
    strlcpy (tmp->name, msg, sizeof(tmp->name));
	tmp->fnct = fnct;

    if (modmsg_list)
        modmsg_list->prev = tmp;

    tmp->next   = modmsg_list;
    modmsg_list   = tmp;

    return;
}

/**********************************************/


void HandleMsg(char *msgtype,char *source, char *dest, char *msg)
{
    ModuleMessage*  mm;
    for (mm = modmsg_list; mm; mm = mm->next)
    {
        if ((strcasecmp (mm->name, msgtype)) == 0)
			mm->fnct(source,dest,msg);
    }
	return;
}

void ev_housekeeping (int argc, char **argv)
{
//    ircd_wallop(s_Guardian->nick, "Performing house cleaning");
    return;
}

void ev_connectuplink (int argc, char **argv)
{
    DelTimedEvent ("Connect Uplink");

    printf ("servsock->sd => %d\n", servsock->sd);

    if (!servsock)
    {
        printf ("No socket connected\n");
        return;
    }

    if (servsock->sd != -1)
    {
        printf ("Socket already connected.\n");
        return;
    }

    if (!Connect(servsock, CfgSettings.uplink, CfgSettings.vhost, CfgSettings.port))
    {
        printf ("Failed to connect uplink\n");
        AddTimedEvent ("Connect Uplink", ev_connectuplink, 60);
        return;
    }

    ircd_connect ();

    return;
}

void ev_ping (int argc, char **argv)
{
	if (!servsock)
		return;

	if (servsock->sd == -1)
		return;

	if (!Uplink)
		return;

	if ((Uplink->lastpong - time(NULL)) > 120)
	{
		uplink_cleanup (NULL, 0);

#ifndef HAVE_PTHREADS
		AddTimedEvent ("Connect Uplink", ev_connectuplink, 60);
#endif

		return;
	}

	send_line ("PING :%s", Uplink->name);

	return;
}


static void SockEventNull(Socket* s, int et)
{
    return;
}

/*************************************************/
/**
 * Add an Error Event to the list - this is used to
 * pickup ERROR codes thrown by the core.
 *
 *  @param errcode - The execption handled by the server
 *  @param fnct    - The routine to handle the exception
 *
 *  @return void
 *
 */


void AddErrHandler (int errcode, void *fnct)
{
    Events      *tmp;
    dlink_node  *dl;

    tmp = (Events *) malloc (sizeof(Events));

    if (!(tmp = (Events *) malloc (sizeof(Events))))
    {
        alog (ERROR, "%s: Unable to malloc()",__PRETTY_FUNCTION__);
        return;
    }

    memcounts.eventssize    += sizeof(tmp);
    memcounts.eventscnt++;

    tmp->handler    = fnct;
    tmp->age        = time(NULL);
    tmp->error      = errcode;

    dl  = dlink_create ();

    dlink_add_tail (tmp, dl, &err_handlers);
    return;
}

/**
 *  NOTE: All errors thrown that dont have a handler
 *        will just fall threw
 */

void catchCoreErr (int errcode, char *errstr)
{
    dlink_node *dl;
    Events *ev;

    DLINK_FOREACH(dl,err_handlers.head)
    {
            ev = dl->data;

            if (ev->error == errcode)
                ev->handler(errcode,(void*) errstr);
            else
                continue;
    }

    return;
}
